#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <time.h>
#include "mockturtle/libmockturtle.h"
#include "libwrtd.h"
#include "libwrtd-private.h"


#define WRTD_DEFAULT_TIMEOUT 1000

static void init (void)
{
        static int initialized;

        if (initialized)
                return;
        initialized = 1;
        trtl_init();
}

enum wrtd_status wrtd_return_error(struct wrtd_dev *dev,
                                   enum wrtd_status error_code,
                                   const char *format, ...)
{
        va_list args;

        if(dev == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        va_start(args, format);
        dev->err = error_code;
        vsnprintf(dev->error_msg, sizeof(dev->error_msg), format, args);
        va_end(args);

        return error_code;
}

enum wrtd_status wrtd_get_error(struct wrtd_dev *wrtd,
                                enum wrtd_status *error_code,
                                int32_t error_description_buffer_size,
                                char *error_description)
{
        enum wrtd_status status;
        int ret;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        char error_message[256];
        memset(error_message, 0, 256);

        status = wrtd_error_message(wrtd, wrtd->err, error_message);
        WRTD_RETURN_IF_ERROR(status);

        int required_buffer_size =
                strlen(error_message) + strlen(wrtd->error_msg) + 1;

        /* If buffer_size is zero, just report on necessary
           buffer size. According to IVI 3.2, section 3.1.2.1. */
        if (error_description_buffer_size == 0)
                return required_buffer_size;

        if(error_code == NULL){
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_NULL_POINTER,
                         "Null pointer passed for function %s, \
                          parameter error_code", __func__);
        }

        if(error_description == NULL){
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_NULL_POINTER,
                         "Null pointer passed for function %s, \
                          parameter error_description", __func__);
        }

        *error_code = wrtd->err;

        /* This violates IVI 3.2, section 3.1.2.1 which dictates to
           ignore overflow if buffer_size is negative. Still, it is
           the safer thing to do.*/
        if (error_description_buffer_size < 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "Invalid value (%d) for function %s, \
                                          parameter error_description_buffer_size",
                                         error_description_buffer_size, __func__);

        ret = snprintf(error_description, error_description_buffer_size,
                       "%s %s", error_message, wrtd->error_msg);
        if (ret < error_description_buffer_size)
                status = WRTD_SUCCESS;
        else
                status = ret + 1;

        wrtd->err = 0;

        memset(wrtd->error_msg, 0, sizeof(wrtd->error_msg));

        return status;
}


enum wrtd_status wrtd_error_message(struct wrtd_dev *wrtd,
                                    enum wrtd_status err_code,
                                    char *err_message)
{

        if(err_message == NULL){
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_NULL_POINTER,
                         "Null pointer passed for function %s, \
                          parameter err_message", __func__);
        }

        switch(err_code){
        case WRTD_SUCCESS:
                snprintf(err_message, 256, "WRTD_SUCCESS");
                break;
        case WRTD_ERROR_INVALID_ATTRIBUTE:
                snprintf(err_message, 256, "WRTD_ERROR_INVALID_ATTRIBUTE");
                break;
        case WRTD_ERROR_ATTR_NOT_WRITEABLE:
                snprintf(err_message, 256, "WRTD_ERROR_ATTR_NOT_WRITEABLE");
                break;
        case WRTD_ERROR_ATTR_NOT_READABLE:
                snprintf(err_message, 256, "WRTD_ERROR_ATTR_NOT_READABLE");
                break;
        case WRTD_ERROR_INVALID_VALUE:
                snprintf(err_message, 256, "WRTD_ERROR_INVALID_VALUE");
                break;
        case WRTD_ERROR_NOT_INITIALIZED:
                snprintf(err_message, 256, "WRTD_ERROR_NOT_INITIALIZED");
                break;
        case WRTD_ERROR_UNKNOWN_CHANNEL_NAME:
                snprintf(err_message, 256, "WRTD_ERROR_UNKNOWN_CHANNEL_NAME");
                break;
        case WRTD_ERROR_OUT_OF_MEMORY:
                snprintf(err_message, 256, "WRTD_ERROR_OUT_OF_MEMORY");
                break;
        case WRTD_ERROR_NULL_POINTER:
                snprintf(err_message, 256, "WRTD_ERROR_NULL_POINTER");
                break;
        case WRTD_ERROR_UNEXPECTED_RESPONSE:
                snprintf(err_message, 256, "WRTD_ERROR_UNEXPECTED_RESPONSE");
                break;
        case WRTD_ERROR_RESOURCE_UNKNOWN:
                snprintf(err_message, 256, "WRTD_ERROR_RESOURCE_UNKNOWN");
                break;
        case WRTD_ERROR_BADLY_FORMED_SELECTOR:
                snprintf(err_message, 256, "WRTD_ERROR_BADLY_FORMED_SELECTOR");
                break;
        case WRTD_ERROR_ALARM_EXISTS:
                snprintf(err_message, 256, "WRTD_ERROR_ALARM_EXISTS");
                break;
        case WRTD_ERROR_ALARM_DOES_NOT_EXIST:
                snprintf(err_message, 256, "WRTD_ERROR_ALARM_DOES_NOT_EXIST");
                break;
        case WRTD_ERROR_VERSION_MISMATCH:
                snprintf(err_message, 256, "WRTD_ERROR_VERSION_MISMATCH");
                break;
        case WRTD_ERROR_INTERNAL:
                snprintf(err_message, 256, "WRTD_ERROR_INTERNAL");
                break;
        case WRTD_ERROR_UNKNOWN_LOG_TYPE:
                snprintf(err_message, 256, "WRTD_ERROR_UNKNOWN_LOG_TYPE");
                break;

        case WRTD_ERROR_RESOURCE_ACTIVE:
                snprintf(err_message, 256, "WRTD_ERROR_RESOURCE_ACTIVE");
                break;
        case WRTD_ERROR_ATTR_GLOBAL:
                snprintf(err_message, 256, "WRTD_ERROR_ATTR_GLOBAL");
                break;
        case WRTD_ERROR_OUT_OF_RESOURCES:
                snprintf(err_message, 256, "WRTD_ERROR_OUT_OF_RESOURCES");
                break;
        case WRTD_ERROR_RULE_EXISTS:
                snprintf(err_message, 256, "WRTD_ERROR_RULE_EXISTS");
                break;
        case WRTD_ERROR_RULE_DOES_NOT_EXIST:
                snprintf(err_message, 256, "WRTD_ERROR_RULE_DOES_NOT_EXIST");
                break;

        default:
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                        "Invalid value (%d) for function %s, \
                                         parameter err_code",
                                         err_code, __func__);
        }

        return WRTD_SUCCESS;
}



enum wrtd_status wrtd_init(const char *resource_name,
                           bool reset,
                           const char *options_str,
                           struct wrtd_dev **dev)
{
        int i;
        struct trtl_dev *trtl;
        const struct trtl_config_rom *cfgrom;
        struct wrtd_config_msg msg;
        enum wrtd_status status;

        /* In case of error...  */
        *dev = NULL;

        /* Initialize (only once).
           FIXME: check errors ? (but there are no errors).  */
        init ();

        /* Resource is MTxxx where xxx is the mock-turtle device id.  */
        if (resource_name[0] == 'M' && resource_name[1] == 'T') {
                unsigned long dev_id;
                char *eptr;
                dev_id = strtoul(resource_name + 2, &eptr, 0);
                if (*eptr != 0) {
                        /* Invalid characters.  */
                        return WRTD_ERROR_RESOURCE_UNKNOWN;
                }
                trtl = trtl_open_by_id(dev_id);
        }
        else if (strncmp(resource_name, "trtl-", 5) == 0) {
                trtl = trtl_open(resource_name);
        }
        else {
                return WRTD_ERROR_RESOURCE_UNKNOWN;
        }
        if (trtl == NULL)
                return WRTD_ERROR_RESOURCE_UNKNOWN;

        struct wrtd_dev *res;

        res = malloc(sizeof(struct wrtd_dev));
        if (res == NULL)
                return WRTD_ERROR_OUT_OF_MEMORY;
        memset(res, 0, sizeof(*res));

        res->trtl = trtl;

        cfgrom = trtl_config_get(res->trtl);

        if (cfgrom->n_cpu > WRTD_MAX_CPUS)
                return wrtd_return_error(res, WRTD_ERROR_INTERNAL,
                                         "%s: Too many CPUs detected.", __func__);

        /* Set HMQ words.  */
        for (i = 0; i < cfgrom->n_cpu; i++)
                if (cfgrom->n_hmq[i] < 0)
                       res->hmq_words[i] = TRTL_CONFIG_ROM_MQ_SIZE_PAYLOAD
                                (cfgrom->hmq[i][0].sizes);

        for (i = 0; i < cfgrom->n_cpu; i++) {
                status = wrtd_msg_get_config(res, i, &msg, __func__);
                WRTD_RETURN_IF_ERROR(status);
                res->root_addr[i] = msg.root_addr;
                /* FIXME: check firmware version ?  */
        }

        res->nbr_cpus = cfgrom->n_cpu;

        *dev = res;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_close(struct wrtd_dev *dev)
{
        if(dev == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        trtl_close(dev->trtl);

        free(dev->alarms);
        //free(dev->rules);
        free(dev);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_reset(struct wrtd_dev *dev)
{
        if(dev == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        return WRTD_SUCCESS;
}


enum wrtd_status wrtd_msg_get_config(struct wrtd_dev *wrtd, unsigned cpu,
                                     struct wrtd_config_msg *res,
                                     const char* caller_func)
{
	struct trtl_msg msg;
	int err;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

	memset(&msg.hdr, 0, sizeof(struct trtl_hmq_header));
	msg.hdr.flags = TRTL_HMQ_HEADER_FLAG_RPC;
	msg.hdr.msg_id = WRTD_ACTION_GET_CONFIG;
        msg.hdr.len = 0;

	err = trtl_msg_sync(wrtd->trtl, cpu, 0, &msg, &msg,
                            WRTD_DEFAULT_TIMEOUT);

	if (err)
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INTERNAL, "%s/%s: %s",
                         caller_func, __func__, trtl_strerror(errno));

	if (msg.hdr.msg_id != WRTD_ACTION_GET_CONFIG
            || msg.hdr.len != 5)
                return wrtd_return_error(wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                         "%s/%s: Unexpected response from device",
                                         caller_func, __func__);


        res->root_addr = msg.data[0];
        res->sync_flag = msg.data[1];
        res->now.seconds = msg.data[2];
        res->now.ns = msg.data[3];
        res->now.frac = msg.data[4];

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_get_sys_time(struct wrtd_dev *wrtd,
                                   struct wrtd_tstamp *time)
{
        struct wrtd_config_msg msg;
        enum wrtd_status status;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        /* Always use CPU #0 (it does exist!).  */
        status = wrtd_msg_get_config(wrtd, 0, &msg, __func__);
        WRTD_RETURN_IF_ERROR(status);

        *time = msg.now;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_fill_roots(struct wrtd_dev *wrtd, const char *caller_func)
{
        enum wrtd_status status;
        unsigned cpu;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        if (wrtd->roots[0].ver_major != 0)
                return WRTD_SUCCESS;

        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++) {
                status = wrtd_msg_readw(wrtd, cpu,
                                        wrtd->root_addr[cpu],
                                        sizeof(struct wrtd_root) / 4,
                                        (uint32_t*)&wrtd->roots[cpu], caller_func);
                if (status != WRTD_SUCCESS)
                        return status;
                if (wrtd->roots[cpu].ver_major != WRTD_VERSION_MAJOR) {
                        status = wrtd_return_error
                                (wrtd, WRTD_ERROR_VERSION_MISMATCH,
                                 "%s/%s: incorrect major version (%d instead of %d)",
                                 caller_func, __func__,
                                 wrtd->roots[cpu].ver_major,
                                 WRTD_VERSION_MAJOR);
                        /* Invalidate root.  */
                        wrtd->roots[cpu].ver_major = 0;
                        return status;
                }
                if (wrtd->roots[cpu].ver_minor != WRTD_VERSION_MINOR) {
                        /* Invalidate root.  */
                        wrtd->roots[cpu].ver_major = 0;
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_VERSION_MISMATCH,
                                 "%s/%s: incorrect minor version (%d instead of %d)",
                                 caller_func, __func__,
                                 wrtd->roots[cpu].ver_minor,
                                 WRTD_VERSION_MINOR);
                }
        }
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_fill_alarms(struct wrtd_dev *wrtd, const char *caller_func)
{
        enum wrtd_status status;
        unsigned cpu;
        unsigned n;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        if (wrtd->alarms != NULL)
                return WRTD_SUCCESS;

        status = wrtd_fill_roots(wrtd, caller_func);
        WRTD_RETURN_IF_ERROR(status);

        /* Count number of alarms.  */
        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++)
                wrtd->nbr_alarms += wrtd->roots[cpu].nbr_alarms;

        /* Allocates.  */
        wrtd->alarms = (struct wrtd_lib_alarm *)
                malloc(wrtd->nbr_alarms * sizeof(struct wrtd_lib_alarm));
        if (wrtd->alarms == NULL)
                return wrtd_return_error(wrtd, WRTD_ERROR_OUT_OF_MEMORY,
                                         "%s/%s: Could not allocate necessary memory.",
                                         caller_func, __func__);
        n = 0;
        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++) {
                unsigned idx;
                for (idx = 0; idx < wrtd->roots[cpu].nbr_alarms; idx++) {
                        status = wrtd_msg_readw
                                (wrtd, cpu,
                                 (wrtd->roots[cpu].alarms_addr
                                  + idx * sizeof(struct wrtd_alarm)),
                                 sizeof(struct wrtd_alarm) / 4,
                                 (uint32_t *)&wrtd->alarms[n].alarm, caller_func);
                        if (status != WRTD_SUCCESS) {
                                free(wrtd->alarms);
                                wrtd->alarms = NULL;
                                return status;
                        }
                        wrtd->alarms[n].cpu = cpu;
                        wrtd->alarms[n].idx = idx;
                        n++;
                }
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_write_alarm(struct wrtd_dev *wrtd, unsigned idx,
                                  const char *caller_func)
{
        enum wrtd_status status;
        struct wrtd_lib_alarm *al;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        al = &wrtd->alarms[idx];

        status = wrtd_msg_writew
                (wrtd, al->cpu,
                 (wrtd->roots[al->cpu].alarms_addr
                  + al->idx * sizeof(struct wrtd_alarm)),
                 sizeof(struct wrtd_alarm) / 4,
                 (uint32_t *)&al->alarm, caller_func);

        return status;
}

enum wrtd_status wrtd_msg_readw(struct wrtd_dev *wrtd, unsigned cpu,
                                uint32_t addr, uint32_t count,
                                uint32_t *dest, const char *caller_func)
{
        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        while (count > 0) {
                uint32_t tlen;
                struct trtl_msg msg;
                int err;
                struct wrtd_io_msg *io_msg;

                /* Partial transaction length.  */
                tlen = wrtd->hmq_words[cpu] - sizeof(struct wrtd_io_msg) / 4;
                if (tlen > count)
                        tlen = count;

                memset(&msg.hdr, 0, sizeof(struct trtl_hmq_header));
                msg.hdr.flags = TRTL_HMQ_HEADER_FLAG_RPC;
                msg.hdr.msg_id = WRTD_ACTION_READW;
                msg.hdr.len = sizeof(struct wrtd_io_msg) / 4;

                io_msg = (struct wrtd_io_msg *)msg.data;
                /* For input: addr + len.  */
                io_msg->addr = addr;
                io_msg->nwords = tlen;

                err = trtl_msg_sync(wrtd->trtl, cpu, 0, &msg, &msg,
                                    WRTD_DEFAULT_TIMEOUT);

                if (err)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_INTERNAL, "%s/%s: %s",
                                 caller_func, __func__, trtl_strerror(errno));

                if (msg.hdr.msg_id != WRTD_ACTION_READW
                    || msg.hdr.len != tlen)
                        return wrtd_return_error(
                                wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                "%s/%s: Unexpected response from device",
                                caller_func, __func__);

                memcpy(dest, msg.data, tlen * sizeof(uint32_t));

                dest += tlen;
                addr += tlen;
                count -= tlen;
        }
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_msg_writew(struct wrtd_dev *wrtd, unsigned cpu,
                                 uint32_t addr, uint32_t count,
                                 const uint32_t *src, const char *caller_func)
{
        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        while (count > 0) {
                uint32_t tlen;
                struct trtl_msg msg;
                int err;
                struct wrtd_io_msg *io_msg;

                /* Partial transaction length.  */
                tlen = wrtd->hmq_words[cpu] - sizeof(struct wrtd_io_msg) / 4;
                if (tlen > count)
                        tlen = count;

                memset(&msg.hdr, 0, sizeof(struct trtl_hmq_header));
                msg.hdr.flags = TRTL_HMQ_HEADER_FLAG_RPC;
                msg.hdr.msg_id = WRTD_ACTION_WRITEW;
                msg.hdr.len = (sizeof(struct wrtd_io_msg) / 4) + tlen;

                io_msg = (struct wrtd_io_msg *)msg.data;
                io_msg->addr = addr;
                io_msg->nwords = tlen;
                memcpy(io_msg->data, src, tlen * sizeof(uint32_t));

                err = trtl_msg_sync(wrtd->trtl, cpu, 0, &msg, &msg,
                                    WRTD_DEFAULT_TIMEOUT);

                if (err)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_INTERNAL, "%s/%s: %s",
                                 caller_func, __func__, trtl_strerror(errno));

                if (msg.hdr.msg_id != WRTD_ACTION_WRITEW
                    || msg.hdr.len != 0)
                        return wrtd_return_error(
                                wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                                "%s/%s: Unexpected response from device",
                                caller_func, __func__);

                src += tlen;
                addr += tlen;
                count -= tlen;
        }
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_write_root_flags(struct wrtd_dev *wrtd, unsigned cpu,
                                       const char *caller_func)
{
        enum wrtd_status status;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        status = wrtd_msg_writew
                (wrtd, cpu,
                 wrtd->root_addr[cpu] + offsetof(struct wrtd_root, log_flags),
                 1, (uint32_t *)&wrtd->roots[cpu].log_flags, caller_func);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_validate_id(struct wrtd_dev *wrtd, const char *rep_cap_id,
                                  const char *caller_func)
{
        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        if ((wrtd_id_null(rep_cap_id)) || (strlen(rep_cap_id) > WRTD_ID_LEN))
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_BADLY_FORMED_SELECTOR,
                         "%s/%s: Badly-formed selector",
                         caller_func, __func__);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_add_alarm(struct wrtd_dev *wrtd,
                                const char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned int i;
        int idx;
        struct wrtd_alarm *alarm;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        /* 1. Validate rep_cap_id.  */
        status = wrtd_validate_id(wrtd, rep_cap_id, __func__);
        WRTD_RETURN_IF_ERROR(status);

        /* 2. Load all alarms.  */
        status = wrtd_fill_alarms(wrtd, __func__);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Find a free slot, check not already used.  */
        idx = -1;
        for (i = 0; i < wrtd->nbr_alarms; i++) {
                if (wrtd_id_eq(wrtd->alarms[i].alarm.event.id, rep_cap_id)) {
                        return wrtd_return_error(wrtd, WRTD_ERROR_ALARM_EXISTS,
                                                 "%s: The alarm already exists",
                                                 __func__);
                }
                if (idx < 0 && wrtd_id_null(wrtd->alarms[i].alarm.event.id))
                        idx = i;
        }
        if (idx == -1)
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_OUT_OF_RESOURCES,
                         "%s: The device has no more resources to allocate.",
                         __func__);

        /* 4. Initialize with default values.  */
        alarm = &wrtd->alarms[idx].alarm;
        /* That should be the default, but... */
        memset(alarm, 0, sizeof(struct wrtd_alarm));
        wrtd_id_copy(alarm->event.id, rep_cap_id);

        /* 5. Write-back.  */
        status = wrtd_write_alarm(wrtd, idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_find_alarm(struct wrtd_dev *wrtd,
                                 const char *rep_cap_id,
                                 unsigned *idx,
                                 const char *caller_func)
{
        enum wrtd_status status;
        unsigned i;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        /* 1. Validate rep_cap_id.  */
        status = wrtd_validate_id(wrtd, rep_cap_id, caller_func);
        WRTD_RETURN_IF_ERROR(status);

        /* 2. Load all alarms.  */
        status = wrtd_fill_alarms(wrtd, caller_func);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Search it.  */
        for (i = 0; i < wrtd->nbr_alarms; i++)
                if (wrtd_id_eq(wrtd->alarms[i].alarm.event.id, rep_cap_id)) {
                        *idx = i;
                        return WRTD_SUCCESS;
                }

        return wrtd_return_error(wrtd, WRTD_ERROR_ALARM_DOES_NOT_EXIST,
                                 "%s/%s: The specified alarm has not been defined",
                                 caller_func, __func__);
}

enum wrtd_status wrtd_remove_alarm(struct wrtd_dev *wrtd,
                                   const char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned idx;
        struct wrtd_alarm *alarm;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        status = wrtd_find_alarm(wrtd, rep_cap_id, &idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_alarm_check_disabled(wrtd, idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        /* Clear it.  */
        alarm = &wrtd->alarms[idx].alarm;
        memset(alarm, 0, sizeof(struct wrtd_alarm));

        /* Write-back.  */
        status = wrtd_write_alarm(wrtd, idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}


enum wrtd_status wrtd_id_copy_buf(struct wrtd_dev *wrtd,
                                  char *dst, int32_t dst_buffer_size,
                                  const char *id, const char *caller_func)
{
        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        /* If buffer_size is zero, just report on necessary
           buffer size. According to IVI 3.2, section 3.1.2.1. */
        if (dst_buffer_size == 0)
                return WRTD_ID_LEN;

        if(dst == NULL){
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_NULL_POINTER,
                         "Null pointer passed for function %s/%s, \
                          parameter dst",
                         caller_func, __func__);
        }

        /* This violates IVI 3.2, section 3.1.2.1 which dictates to
           ignore overflow if buffer_size is negative. Still, it is
           the safer thing to do.*/
        if (dst_buffer_size < 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "Invalid value (%d) for function %s/%s, \
                                          parameter dst_buffer_size",
                                         dst_buffer_size, caller_func, __func__);

        memset(dst, 0, dst_buffer_size);

        int ret = snprintf(dst, dst_buffer_size, "%s", id);
        if (ret < dst_buffer_size)
                return WRTD_SUCCESS;
        else
                return ret + 1;
}


enum wrtd_status wrtd_get_alarm_id(struct wrtd_dev *wrtd,
                                   int32_t index,
                                   int32_t name_buffer_size,
                                   char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned i;
        unsigned n;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        /* Load all alarms.  */
        status = wrtd_fill_alarms(wrtd, __func__);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Search it.  */
        for (n = 0, i = 0; i < wrtd->nbr_alarms; i++) {
                const char *aid = wrtd->alarms[i].alarm.event.id;
                if (wrtd_id_null(aid))
                        continue;
                if (n == index) {
                        status = wrtd_id_copy_buf
                                (wrtd, rep_cap_id, name_buffer_size, aid, __func__);
                        WRTD_RETURN_IF_ERROR(status);
                        return WRTD_SUCCESS;
                }
                n++;
        }

        return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                 "Invalid value (%d) for function %s, \
                                  parameter index.", index, __func__);
}

enum wrtd_status wrtd_log_read(struct wrtd_dev *wrtd,
                               struct wrtd_log_entry *log,
                               const char *caller_func)
{
	struct trtl_msg msg;
	struct polltrtl p[WRTD_MAX_CPUS];
	int ret, i;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

	for (i = 0; i < wrtd->nbr_cpus; ++i) {
		p[i].trtl = wrtd->trtl;
		p[i].idx_hmq = WRTD_HMQ;
                p[i].idx_cpu = i;
		p[i].events = POLLIN;
	}

	/* Clean up errno to be able to distinguish between error cases and
	   normal behaviour when the function return less messages
	   than expected */
        errno = 0;

        ret = trtl_msg_poll(p, wrtd->nbr_cpus, 0);
        if (ret == 0) {
                // Nothing to read
                log->type = WRTD_LOG_MSG_EV_NONE;
                return WRTD_SUCCESS;
        }
        if (ret < 0)
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_INTERNAL, "%s/%s: %s",
                         caller_func, __func__, trtl_strerror(errno));

        for (i = 0; i < wrtd->nbr_cpus; ++i) {
                if (!(p[i].revents & POLLIN))
                        continue;
                ret = trtl_msg_async_recv(wrtd->trtl, i, WRTD_HMQ, &msg, 1);
                if (ret <= 0)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_INTERNAL, "%s/%s: %s",
                                 caller_func, __func__, trtl_strerror(errno));

                memcpy(log, msg.data, sizeof(struct wrtd_log_entry));
                return WRTD_SUCCESS;
        }
        /* Should not happen.  */
        return wrtd_return_error(
                wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                "%s/%s: Unexpected response from device",
                caller_func, __func__);
}

enum wrtd_status wrtd_get_next_event_log_entry(struct wrtd_dev *wrtd,
                                               int32_t log_entry_buffer_size,
                                               char *log_entry)
{
        struct wrtd_log_entry log;
        enum wrtd_status status;
        struct tm *tm;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        /* Do not retrieve entry from log if buffer_size
           is zero. Just report on necessary buffer size.
           According to IVI 3.2, section 3.1.2.1 */
        if (log_entry_buffer_size == 0)
                return WRTD_LOG_ENTRY_SIZE;

        if(log_entry == NULL){
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_NULL_POINTER,
                         "Null pointer passed for function %s, \
                          parameter log_entry", __func__);
        }

        status = wrtd_log_read(wrtd, &log, __func__);
        WRTD_RETURN_IF_ERROR(status);

        memset(log_entry, 0, log_entry_buffer_size);

        /* Nothing to read */
        if (log.type == WRTD_LOG_MSG_EV_NONE)
                return WRTD_SUCCESS;

        /* This violates IVI 3.2, section 3.1.2.1 which dictates to
           ignore overflow if buffer_size is negative. Still, it is
           the safer thing to do.*/
        if (log_entry_buffer_size < 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                         "Invalid value (%d) for function %s, \
                                          parameter log_entry_buffer_size",
                                         log_entry_buffer_size, __func__);

        int count = log_entry_buffer_size;

        char *sptr = log_entry;

        tm = gmtime((const time_t *)&log.ts.seconds);
        if (tm == NULL)
                return wrtd_return_error(
                        wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                        "%s: Unexpected response from device",
                        __func__);

        /* Needs 20 characters */
        strftime(sptr, count, "%F,%T.", tm);
        count -= 20;
        if (count < 0)
                return WRTD_LOG_ENTRY_SIZE;
        sptr  += 20;

        /* Needs 16 characters */
        snprintf(sptr, count, "%03lu.%03lu.%03lu+%03d|",
                 (log.ts.ns / (1000L * 1000)),
                 (log.ts.ns / 1000L) % 1000,
                 log.ts.ns % 1000UL,
                 log.ts.frac >> (32 - 9));
        count -= 16;
        if (count < 0)
                return WRTD_LOG_ENTRY_SIZE;
        sptr  += 16;

        tm = gmtime((const time_t *)&log.event.ts.seconds);
        if (tm == NULL)
                return wrtd_return_error(
                        wrtd, WRTD_ERROR_UNEXPECTED_RESPONSE,
                        "%s: Unexpected response from device",
                        __func__);

        /* Needs 20 characters */
        strftime(sptr, count, "%F,%T.", tm);
        count -= 20;
        if (count < 0)
                return WRTD_LOG_ENTRY_SIZE;
        sptr  += 20;

        /* Needs 16 characters */
        snprintf(sptr, count, "%03lu.%03lu.%03lu+%03d|",
                 (log.event.ts.ns / (1000L * 1000)),
                 (log.event.ts.ns / 1000L) % 1000,
                 log.event.ts.ns % 1000UL,
                 log.event.ts.frac >> (32 - 9));
        count -= 16;
        if (count < 0)
                return WRTD_LOG_ENTRY_SIZE;
        sptr  += 16;

        /* Needs 18 characters */
        switch(log.type) {
        case WRTD_LOG_MSG_EV_GENERATED:
                if (log.reason == WRTD_LOG_GENERATED_ALARM)
                        snprintf(sptr, count, "GENERATED|ALARM   ");
                else if ((log.reason >= WRTD_LOG_GENERATED_DEVICE_0) &&
                         (log.reason <= WRTD_LOG_GENERATED_DEVICE_7) &&
                         (log.reason % 8 == 0))
                        snprintf(sptr, count, "GENERATED|DEVICE_%d",
                                 (log.reason - 8) / 8);
                else
                        return wrtd_return_error(
                                wrtd, WRTD_ERROR_UNKNOWN_LOG_TYPE,
                                "%s: Unknown log reason (%#x)",
                                __func__, log.reason);
                break;
        case WRTD_LOG_MSG_EV_CONSUMED:
                if (log.reason == WRTD_LOG_CONSUMED_START)
                        snprintf(sptr, count, "CONSUMED |START   ");
                else if (log.reason == WRTD_LOG_CONSUMED_DONE)
                        snprintf(sptr, count, "CONSUMED |DONE    ");
                else
                        return wrtd_return_error(
                                wrtd, WRTD_ERROR_UNKNOWN_LOG_TYPE,
                                "%s: Unknown log reason (%#x)",
                                __func__, log.reason);
                break;
        case WRTD_LOG_MSG_EV_DISCARDED:
                if (log.reason == WRTD_LOG_DISCARD_NO_SYNC)
                        snprintf(sptr, count, "DISCARDED|NO SYNC ");
                else if (log.reason == WRTD_LOG_DISCARD_HOLDOFF)
                        snprintf(sptr, count, "DISCARDED|HOLD OFF");
                else if (log.reason == WRTD_LOG_DISCARD_TIMEOUT)
                        snprintf(sptr, count, "DISCARDED|TIME OUT");
                else if (log.reason == WRTD_LOG_DISCARD_OVERFLOW)
                        snprintf(sptr, count, "DISCARDED|OVERFLOW");
                else
                        return wrtd_return_error(
                                wrtd, WRTD_ERROR_UNKNOWN_LOG_TYPE,
                                "%s: Unknown log reason (%#x)",
                                __func__, log.reason);
                break;
        case WRTD_LOG_MSG_EV_NETWORK:
                if (log.reason == WRTD_LOG_NETWORK_TX)
                        snprintf(sptr, count, "NETWORK  |TX      ");
                else if (log.reason == WRTD_LOG_NETWORK_RX)
                        snprintf(sptr, count, "NETWORK  |RX      ");
                else
                        return wrtd_return_error(
                                wrtd, WRTD_ERROR_UNKNOWN_LOG_TYPE,
                                "%s: Unknown log reason (%#x)",
                                __func__, log.reason);
                break;
        default:
                return wrtd_return_error(
                        wrtd, WRTD_ERROR_UNKNOWN_LOG_TYPE,
                        "%s: Unknown log type (%#x)",
                        __func__, log.type);
                return WRTD_ERROR_UNKNOWN_LOG_TYPE;
        }

        /* Last check, therefore also account for null termination */
        count -= 19;
        if (count < 0)
                return WRTD_LOG_ENTRY_SIZE;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_clear_event_log_entries(struct wrtd_dev *wrtd)
{
        int ret, i;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        for (i = 0; i < wrtd->nbr_cpus; ++i) {
                ret = trtl_hmq_flush(wrtd->trtl, i, WRTD_HMQ);
                if (ret < 0)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_INTERNAL,
                                 "%s: %s", __func__, trtl_strerror(errno));
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_alloc_rules(struct wrtd_dev *wrtd,
                                  struct wrtd_lib_rule **rulesp,
                                  const char *caller_func)
{
        struct wrtd_lib_rule *rules;
        unsigned n;
        unsigned cpu;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        rules = (struct wrtd_lib_rule *)
                calloc(wrtd->nbr_rules, sizeof(struct wrtd_lib_rule));
        if (rules == NULL)
                return wrtd_return_error(wrtd, WRTD_ERROR_OUT_OF_MEMORY,
                                         "%s/%s: Could not allocate necessary memory.",
                                         caller_func, __func__);

        n = 0;
        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++) {
                unsigned nbr_rules = wrtd->roots[cpu].nbr_rules;
                unsigned idx;
                for (idx = 0; idx < nbr_rules; idx++) {
                        rules[n].cpu = cpu;
                        rules[n].local_idx = idx;
                        n++;
                }
        }
        assert(n == wrtd->nbr_rules);
        *rulesp = rules;

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_fill_rules(struct wrtd_dev *wrtd, const char *caller_func)
{
        enum wrtd_status status;
        unsigned i;
        unsigned cpu;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        if (wrtd->rules != NULL)
                return WRTD_SUCCESS;

        status = wrtd_fill_roots(wrtd, caller_func);
        WRTD_RETURN_IF_ERROR(status);

        /* Count number of rules.  */
        assert(wrtd->nbr_rules == 0);
        for (cpu = 0; cpu < wrtd->nbr_cpus; cpu++)
                wrtd->nbr_rules += wrtd->roots[cpu].nbr_rules;

        /* Allocates.  */
        status = wrtd_alloc_rules(wrtd, &wrtd->rules, caller_func);
        WRTD_RETURN_IF_ERROR(status);

        /* Read rules from cpus.  */
        for (i = 0; i < wrtd->nbr_rules; i++) {
                unsigned cpu = wrtd->rules[i].cpu;
                unsigned int addr;

                addr = (wrtd->roots[cpu].rules_addr
                        + wrtd->rules[i].local_idx * sizeof(struct wrtd_rule));
                status = wrtd_msg_readw
                        (wrtd, cpu, addr, sizeof(struct wrtd_rule) / 4,
                         (uint32_t *)&wrtd->rules[i].rule, caller_func);
                if (status != WRTD_SUCCESS) {
                        free(wrtd->rules);
                        wrtd->rules = NULL;
                        return status;
                }
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_write_rule(struct wrtd_dev *wrtd, unsigned idx,
                                 const char *caller_func)
{
        enum wrtd_status status;
        struct wrtd_lib_rule *rule;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        rule = &wrtd->rules[idx];

        status = wrtd_msg_writew
                (wrtd, rule->cpu,
                 (wrtd->roots[rule->cpu].rules_addr
                  + rule->local_idx * sizeof(struct wrtd_rule)),
                 sizeof(struct wrtd_rule) / 4,
                 (uint32_t *)&rule->rule, caller_func);

        return status;
}

enum wrtd_status wrtd_add_rule(struct wrtd_dev *wrtd,
                                const char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned int i;
        int idx;
        struct wrtd_rule *rule;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        /* 1. Validate rep_cap_id.  */
        status = wrtd_validate_id(wrtd, rep_cap_id, __func__);
        WRTD_RETURN_IF_ERROR(status);

        /* 2. Load all rules.  */
        status = wrtd_fill_rules(wrtd, __func__);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Find a free slot, check not already used.  */
        idx = -1;
        for (i = 0; i < wrtd->nbr_rules; i++) {
                if (wrtd_id_eq(wrtd->rules[i].rule.conf.id, rep_cap_id)) {
                        return wrtd_return_error(wrtd, WRTD_ERROR_RULE_EXISTS,
                                                 "%s: The rule already exists",
                                                 __func__);
                }
                if (idx < 0
                    && wrtd_id_null(wrtd->rules[i].rule.conf.id))
                        idx = i;
        }
        if (idx == -1)
                return wrtd_return_error
                        (wrtd, WRTD_ERROR_OUT_OF_RESOURCES,
                         "%s: The device has no more resources to allocate.",
                         __func__);

        /* 4. Initialize with default values.  */
        rule = &wrtd->rules[idx].rule;
        /* That should be the default, but... */
        memset(rule, 0, sizeof(struct wrtd_rule));
        wrtd_id_copy(rule->conf.id, rep_cap_id);
        rule->conf.send_late = true;

        /* 5. Write-back.  */
        status = wrtd_write_rule(wrtd, idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_find_rule(struct wrtd_dev *wrtd,
                                const char *rep_cap_id,
                                unsigned *idx,
                                const char *caller_func)
{
        enum wrtd_status status;
        unsigned i;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        /* 1. Validate rep_cap_id.  */
        status = wrtd_validate_id(wrtd, rep_cap_id, caller_func);
        WRTD_RETURN_IF_ERROR(status);

        /* 2. Load all rules.  */
        status = wrtd_fill_rules(wrtd, caller_func);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Search it.  */
        for (i = 0; i < wrtd->nbr_rules; i++)
                if (wrtd_id_eq(wrtd->rules[i].rule.conf.id, rep_cap_id)) {
                        *idx = i;
                        return WRTD_SUCCESS;
                }

        return wrtd_return_error(wrtd, WRTD_ERROR_RULE_DOES_NOT_EXIST,
                                 "%s/%s: The specified rule has not been defined",
                                 caller_func, __func__);
}

enum wrtd_status wrtd_get_rule_id(struct wrtd_dev *wrtd,
                                  int32_t index,
                                  int32_t name_buffer_size,
                                  char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned i;
        unsigned n;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        /* Load all rules.  */
        status = wrtd_fill_rules(wrtd, __func__);
        WRTD_RETURN_IF_ERROR(status);

        /* 3. Search it.  */
        for (n = 0, i = 0; i < wrtd->nbr_rules; i++) {
                const char *aid = wrtd->rules[i].rule.conf.id;
                if (wrtd_id_null(aid))
                        continue;
                if (n == index) {
                        status = wrtd_id_copy_buf
                                (wrtd, rep_cap_id, name_buffer_size, aid, __func__);
                        WRTD_RETURN_IF_ERROR(status);
                        return WRTD_SUCCESS;
                }
                n++;
        }

        return wrtd_return_error(wrtd, WRTD_ERROR_INVALID_VALUE,
                                 "Invalid value (%d) for function %s, \
                                  parameter index.", index, __func__);
}

enum wrtd_status wrtd_remove_rule(struct wrtd_dev *wrtd,
                                  const char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned idx;
        struct wrtd_rule *rule;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        /* Clear it.  */
        rule = &wrtd->rules[idx].rule;
        memset(rule, 0, sizeof(struct wrtd_rule));

        /* Write-back.  */
        status = wrtd_write_rule(wrtd, idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_remove_all_rules(struct wrtd_dev *wrtd)
{
        enum wrtd_status status;
        unsigned idx;
        struct wrtd_rule *rule;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        status = wrtd_fill_rules(wrtd, __func__);
        WRTD_RETURN_IF_ERROR(status);

        for (idx = 0; idx < wrtd->nbr_rules; idx++) {
                status = wrtd_rule_check_disabled(wrtd, idx, __func__);
                WRTD_RETURN_IF_ERROR(status);
        }

        for (idx = 0; idx < wrtd->nbr_rules; idx++) {
                rule = &wrtd->rules[idx].rule;

                /* Clear it.  */
                memset(rule, 0, sizeof(struct wrtd_rule));

                /* Write-back.  */
                status = wrtd_write_rule(wrtd, idx, __func__);
                WRTD_RETURN_IF_ERROR(status);

        }

        return WRTD_SUCCESS;
}


extern enum wrtd_status wrtd_remove_all_alarms(struct wrtd_dev *wrtd)
{
        enum wrtd_status status;
        unsigned idx;
        struct wrtd_alarm *alarm;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        status = wrtd_fill_alarms(wrtd, __func__);
        WRTD_RETURN_IF_ERROR(status);

        for (idx = 0; idx < wrtd->nbr_alarms; idx++) {
                status = wrtd_alarm_check_disabled(wrtd, idx, __func__);
                WRTD_RETURN_IF_ERROR(status);
        }

        for (idx = 0; idx < wrtd->nbr_alarms; idx++) {
                /* Clear it.  */
                alarm = &wrtd->alarms[idx].alarm;
                memset(alarm, 0, sizeof(struct wrtd_alarm));

                /* Write-back.  */
                status = wrtd_write_alarm(wrtd, idx, __func__);
                WRTD_RETURN_IF_ERROR(status);
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_disable_all_rules(struct wrtd_dev *wrtd)
{
        unsigned idx;
        enum wrtd_status status;
        struct wrtd_lib_rule *r;
        int changed = 0;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        status = wrtd_fill_rules(wrtd, __func__);
        WRTD_RETURN_IF_ERROR(status);

        for (idx = 0; idx < wrtd->nbr_rules; idx++) {
                r = &wrtd->rules[idx];
                if (r->rule.conf.enabled != 0) {
                        r->rule.conf.enabled = 0;
                        changed = 1;
                }
        }
        if (changed) {
                status = wrtd_reconfigure(wrtd, __func__);
                WRTD_RETURN_IF_ERROR(status);
        }
        return WRTD_SUCCESS;
}


enum wrtd_status wrtd_disable_all_alarms(struct wrtd_dev *wrtd)
{
        unsigned idx;
        enum wrtd_status status;
        struct wrtd_alarm *alarm;
        int changed = 0;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        status = wrtd_fill_alarms(wrtd, __func__);
        WRTD_RETURN_IF_ERROR(status);

        for (idx = 0; idx < wrtd->nbr_alarms; idx++) {
                alarm = &wrtd->alarms[idx].alarm;
                if(alarm->enabled != 0) {
                        alarm->enabled = 0;
                        changed = 1;
                }
        }

        if (changed) {
                status = wrtd_reconfigure(wrtd, __func__);
                WRTD_RETURN_IF_ERROR(status);
        }
        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_reset_rule_stats(struct wrtd_dev *wrtd,
                                       const char *rep_cap_id)
{
        enum wrtd_status status;
        unsigned idx;
        struct wrtd_rule *rule;

        if(wrtd == NULL){
                return WRTD_ERROR_NOT_INITIALIZED;
        }

        status = wrtd_find_rule(wrtd, rep_cap_id, &idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_rule_check_disabled(wrtd, idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        rule = &wrtd->rules[idx].rule;

        /* Clear stats.  */
        memset(&rule->stat, 0, sizeof(struct wrtd_rule_stats));

        /* Write-back.  */
        status = wrtd_write_rule(wrtd, idx, __func__);
        WRTD_RETURN_IF_ERROR(status);

        return WRTD_SUCCESS;
}
