#ifndef __LIBWRTD_PRIVATE__H__
#define __LIBWRTD_PRIVATE__H__

#include <strings.h>
#include "libwrtd.h"
#include "mockturtle/libmockturtle.h"

#define WRTD_RETURN_IF_ERROR(status) if(status != WRTD_SUCCESS) return status

/* A complete log entry needs 35+1+35+1+9+1+8=90 + 1 for null termination.
   1st field: log tstamp in YYYY-MM-DD,hh:mm:ss.xxx.yyy.zzz+fff format.
   2nd field: event tstamp in YYYY-MM-DD,hh:mm:ss.xxx.yyy.zzz+fff format.
   3rd field: log type, padded with spaces to 9 characters.
   4th field: log reason, padded with spaces to 8 characters. */
#define WRTD_LOG_ENTRY_SIZE 91

/* Add nanoseconds to an existing wrtd timestamp */
static inline void wrtd_ts_add_ns(struct wrtd_tstamp *ts, uint32_t ns)
{
        ts->ns += ns;
        while (ts->ns >= 1000000000) {
                ts->ns -= 1000000000;
                ts->seconds++;
        }
}

/* Alarm when in user space.  */
struct wrtd_lib_alarm {
        /* CPU on which the alarm is.  */
        unsigned int cpu;
        /* Alarm index for the cpu.  */
        unsigned int idx;
        /* Alarm data.  */
        struct wrtd_alarm alarm;
};

struct wrtd_lib_rule {
        /* MT CPU on which the rule is placed.  */
        unsigned int cpu;
        /* Rule index on the local MT cpu.  */
        unsigned int local_idx;
        /* Rule.  */
        struct wrtd_rule rule;
};

struct wrtd_dev {
        /* Associated mock-turtle device.  */
        struct trtl_dev *trtl;

        /* Last error.  */
        enum wrtd_status err;

        /* Long error message.  */
        char error_msg[256];

        /* Number of cpus. Valid if > 0. */
        uint32_t nbr_cpus;

        /* Per-cpu hmq length (from MT config) - this is needed for
           readw/writew. */
        uint32_t hmq_words[WRTD_MAX_CPUS];

        /* Per-cpu root address.  */
        uint32_t root_addr[WRTD_MAX_CPUS];

        /* Roots.
           Valid only when versions are not 0.  */
        struct wrtd_root roots[WRTD_MAX_CPUS];

        /* Rules.  */
        uint32_t nbr_rules;
        struct wrtd_lib_rule *rules;

        /* Alarms.  */
        uint32_t nbr_alarms;
        struct wrtd_lib_alarm *alarms;
};

/* Save the error in DEV and return the error code.  */
extern enum wrtd_status wrtd_return_error(struct wrtd_dev *dev,
                                          enum wrtd_status error_code,
                                          const char *format, ...)
        __attribute__((format(printf, 3, 4)));

/* Do the get_config action. */
enum wrtd_status wrtd_msg_get_config(struct wrtd_dev *wrtd, unsigned cpu,
                                     struct wrtd_config_msg *res,
                                     const char *caller_func);

/* Read words from CPU space.  */
enum wrtd_status wrtd_msg_readw(struct wrtd_dev *wrtd, unsigned cpu,
                                uint32_t addr, uint32_t count,
                                uint32_t *dest,
                                const char *caller_func);

/* Write words to CPU space.  */
enum wrtd_status wrtd_msg_writew(struct wrtd_dev *wrtd, unsigned cpu,
                                 uint32_t addr, uint32_t count,
                                 const uint32_t *src,
                                 const char *caller_func);

/* Check if rep_cap_id is valid (not null, length). */
enum wrtd_status wrtd_validate_id(struct wrtd_dev *wrtd,
                                  const char *rep_cap_id,
                                  const char *caller_func);

/* Ensure the roots have been fetched from the CPUs.  */
enum wrtd_status wrtd_fill_roots(struct wrtd_dev *wrtd,
                                 const char *caller_func);

enum wrtd_status wrtd_write_root_flags(struct wrtd_dev *wrtd, unsigned cpu,
                                       const char *caller_func);

/* Ensure the alarms have been fetched from the CPUs.  */
enum wrtd_status wrtd_fill_alarms(struct wrtd_dev *wrtd,
                                  const char *caller_func);

/* Write back alarm idx. */
enum wrtd_status wrtd_write_alarm(struct wrtd_dev *wrtd, unsigned idx,
                                  const char *caller_func);

/* Find alarm index by id. */
enum wrtd_status wrtd_find_alarm(struct wrtd_dev *wrtd,
                                 const char *rep_cap_id,
                                 unsigned *idx,
                                 const char *caller_func);

/* Find rule index by id. */
enum wrtd_status wrtd_find_rule(struct wrtd_dev *wrtd,
                                const char *rep_cap_id,
                                unsigned *idx,
                                const char *caller_func);

/* Internal function: allocate rules and set cpu,idx.  */
enum wrtd_status wrtd_alloc_rules(struct wrtd_dev *wrtd,
                                  struct wrtd_lib_rule **rulesp,
                                  const char *caller_func);

/* Write-back rule. */
enum wrtd_status wrtd_write_rule(struct wrtd_dev *wrtd, unsigned idx,
                                 const char *caller_func);

/* Ensure the rules have been fetched from the CPUs.  */
enum wrtd_status wrtd_fill_rules(struct wrtd_dev *wrtd,
                                 const char *caller_func);

/* Count number of rules.  */
enum wrtd_status wrtd_attr_get_rule_count(struct wrtd_dev *wrtd,
                                          int32_t *value);

/* Reconfigure route tables and write-back rules.  */
enum wrtd_status wrtd_reconfigure(struct wrtd_dev *wrtd, const char *caller_func);

static inline bool wrtd_id_eq(const char *l, const char *r)
{
        return strncmp(l, r, WRTD_ID_LEN) == 0;
}

static inline bool wrtd_id_null(const char *id)
{
        return id[0] == 0;
}

static inline void wrtd_id_copy(char *dst, const char *src)
{
        strncpy(dst, src, WRTD_ID_LEN);
}

static inline enum wrtd_status wrtd_alarm_check_disabled(struct wrtd_dev *wrtd, unsigned idx,
                                                         const char *caller_func)
{
        if (wrtd->alarms[idx].alarm.enabled)
                return wrtd_return_error(wrtd, WRTD_ERROR_RESOURCE_ACTIVE,
                                         "%s/%s: Resource is enabled.",
                                         caller_func, __func__);
        return WRTD_SUCCESS;
}

static inline enum wrtd_status wrtd_rule_check_disabled(struct wrtd_dev *wrtd, unsigned idx,
                                                        const char *caller_func)
{
        if (wrtd->rules[idx].rule.conf.enabled)
                return wrtd_return_error(wrtd, WRTD_ERROR_RESOURCE_ACTIVE,
                                         "%s/%s: Resource is enabled.",
                                         caller_func, __func__);
        return WRTD_SUCCESS;
}

/* Copy ID to DST of length BUF_SIZE.  Check length, pad with null.  */
enum wrtd_status wrtd_id_copy_buf(struct wrtd_dev *wrtd,
                                  char *dst, int32_t buf_size, const char *id,
                                  const char *caller_func);

#endif /* __LIBWRTD_PRIVATE__H__ */
