#ifndef __WRTD_LIB_H__
#define __WRTD_LIB_H__
/** @file libwrtd-ng.h */

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stdbool.h>
#include "wrtd-common.h"

// NOTE: corresponds roughly to IVI ViSession
struct wrtd_dev;

/**
 * @enum wrtd_status
 * White Rabbit Trigger Distribution error codes.
 * Names and values inspired by IVI 3.2 and IVI 3.15.
 */
enum wrtd_status {
        /** No error. */
        WRTD_SUCCESS = 0,

        /** Same as IVI_INHERENT_ERROR_BASE */
        __WRTD_ERROR_BASE = 0xBFFA0000,
        /** Invalid/unknown attribute. */
        WRTD_ERROR_INVALID_ATTRIBUTE     = __WRTD_ERROR_BASE + 0x0C,
        /** Attempt to write a read-only attribute. */
        WRTD_ERROR_ATTR_NOT_WRITEABLE    = __WRTD_ERROR_BASE + 0x0D,
        /** Attempt to read a write-only attribute. */
        WRTD_ERROR_ATTR_NOT_READABLE     = __WRTD_ERROR_BASE + 0x0E,
        /** Invalid value provided. */
        WRTD_ERROR_INVALID_VALUE         = __WRTD_ERROR_BASE + 0x10,
        /** Device not initialized. */
        WRTD_ERROR_NOT_INITIALIZED       = __WRTD_ERROR_BASE + 0x1D,
        /** Unknown channel name. */
        WRTD_ERROR_UNKNOWN_CHANNEL_NAME  = __WRTD_ERROR_BASE + 0x20,
        /** Device out of memory. */
        WRTD_ERROR_OUT_OF_MEMORY         = __WRTD_ERROR_BASE + 0x56,
        /** Null pointer. */
        WRTD_ERROR_NULL_POINTER          = __WRTD_ERROR_BASE + 0x58,
        /** Unexpected response. */
        WRTD_ERROR_UNEXPECTED_RESPONSE   = __WRTD_ERROR_BASE + 0x59,
        /** Unknown resource. */
        WRTD_ERROR_RESOURCE_UNKNOWN      = __WRTD_ERROR_BASE + 0x60,
        /** Incorrect repeated capability selector. */
        WRTD_ERROR_BADLY_FORMED_SELECTOR = __WRTD_ERROR_BASE + 0x66,

        /** Same as IVI_LXISYNC_ERROR_BASE */
        __WRTD_LXISYNC_ERROR_BASE = 0xBFFA3000,
        /** The alarm already exists. */
        WRTD_ERROR_ALARM_EXISTS          = __WRTD_LXISYNC_ERROR_BASE + 0x07,
        /** The specified alarm has not been defined. */
        WRTD_ERROR_ALARM_DOES_NOT_EXIST  = __WRTD_LXISYNC_ERROR_BASE + 0x08,

        /** Same as IVI_VENDOR_SPECIFIC_ERROR_BASE */
        __WRTD_SPECIFIC_ERROR_BASE = 0xBFFA6000,
        /** Version mismatch */
        WRTD_ERROR_VERSION_MISMATCH       = __WRTD_SPECIFIC_ERROR_BASE + 0x00,
        /** Internal error */
        WRTD_ERROR_INTERNAL               = __WRTD_SPECIFIC_ERROR_BASE + 0x01,
        /** Unknown event log type/reason */
        WRTD_ERROR_UNKNOWN_LOG_TYPE       = __WRTD_SPECIFIC_ERROR_BASE + 0x02,
        /** Resource is not disabled and cannot be changed.  */
        WRTD_ERROR_RESOURCE_ACTIVE        = __WRTD_SPECIFIC_ERROR_BASE + 0x03,
        /** Attempt to access a global attribute without
            using the global attribute selector */
        WRTD_ERROR_ATTR_GLOBAL            = __WRTD_SPECIFIC_ERROR_BASE + 0x04,
        /** The device has no more resources to allocate. */
        WRTD_ERROR_OUT_OF_RESOURCES       = __WRTD_SPECIFIC_ERROR_BASE + 0x05,
        /** The rule already exists. */
        WRTD_ERROR_RULE_EXISTS            = __WRTD_SPECIFIC_ERROR_BASE + 0x06,
        /** The specified rule has not been defined. */
        WRTD_ERROR_RULE_DOES_NOT_EXIST    = __WRTD_SPECIFIC_ERROR_BASE + 0x07,

        /** always last entry in this enum */
        __WRTD_ERROR_MAX_NUMBER,
};

/**
 * @enum wrtd_attr
 * White Rabbit Trigger Distribution attribute ID definions
 */
enum wrtd_attr {
	__WRTD_ATTR_BASE = 950000,

	WRTD_MAJOR_VERSION, // RO, int32
	WRTD_MINOR_VERSION, // RO, int32
	WRTD_ATTR_EVENT_LOG_EMPTY, // RO, bool
	WRTD_ATTR_EVENT_LOG_ENABLED, //RW, bool
	// TODO: WRTD_ATTR_IS_TIME_MASTER, // RO, bool
	WRTD_ATTR_IS_TIME_SYNCHRONIZED, // RO, bool

        // Number of alarms (global attribute).
	WRTD_ATTR_ALARM_COUNT, //RO, int32

        // Enable/disable an alarm.
	WRTD_ATTR_ALARM_ENABLED, //RW, bool
	// Specifies at what time to trigger the alarm
	WRTD_ATTR_ALARM_TIME, //RW, tstamp
	// Specifies the alarm period. 0 means no repetitions.
	WRTD_ATTR_ALARM_PERIOD, //RW, tstamp
	// Specifies the number of times the alarm will occur at the period specified by
	// WRTD_ATTR_ALARM_PERIOD. 0 means infinite. 1 means exactly one alarm will occur.
	WRTD_ATTR_ALARM_REPEAT_COUNT, //RW, int32

        // Number of rules (global attribute).
	WRTD_ATTR_RULE_COUNT, //RO, int32

        // True if the rule is enabled (ie active).
	WRTD_ATTR_RULE_ENABLED, //RW, bool
        // Number of time the rule will apply.  0 means infinite.  When read,
        // return the remaining number.
	WRTD_ATTR_RULE_REPEAT_COUNT, //RW, int32
        // Source of the event.
	WRTD_ATTR_RULE_SOURCE, //RW, string
        // Destination of the event.
	WRTD_ATTR_RULE_DESTINATION, //RW, string
        // If true, events from the past are still handled; otherwise they
        // are discarded.
	WRTD_ATTR_RULE_SEND_LATE, //RW, bool
        // Delay the event.
	WRTD_ATTR_RULE_DELAY, //RW, tstamp
        // Discard new events between the last one to this value.
	WRTD_ATTR_RULE_HOLDOFF, //RW, tstamp
        // Realign the timestamp to a multiple of period and then add
        // factor * period. This is done after the delay.
	WRTD_ATTR_RULE_RESYNC_PERIOD, //RW, tstamp
	WRTD_ATTR_RULE_RESYNC_FACTOR, //RW, int32

        /* Statistics */
	WRTD_ATTR_STAT_RULE_RX_EVENTS, //RO, int32
	WRTD_ATTR_STAT_RULE_RX_LAST, //RO, tstamp
	WRTD_ATTR_STAT_RULE_TX_EVENTS, //RO, int32
	WRTD_ATTR_STAT_RULE_TX_LAST, //RO, tstamp
	WRTD_ATTR_STAT_RULE_MISSED_LATE, //RO, int32
	WRTD_ATTR_STAT_RULE_MISSED_HOLDOFF, //RO, int32
	WRTD_ATTR_STAT_RULE_MISSED_NOSYNC, //RO, int32
	WRTD_ATTR_STAT_RULE_MISSED_OVERFLOW, //RO, int32
	WRTD_ATTR_STAT_RULE_MISSED_LAST, //RO, tstamp
	WRTD_ATTR_STAT_RULE_RX_LATENCY_MIN, //RO, tstamp
	WRTD_ATTR_STAT_RULE_RX_LATENCY_MAX, //RO, tstamp
	WRTD_ATTR_STAT_RULE_RX_LATENCY_AVG, //RO, tstamp
	WRTD_ATTR_STAT_RULE_RX_PERIOD_AVG, //RO, tstamp

	__WRTD_ATTR_MAX_NUMBER,
};

/**
 * A repeated capability identifier for global attributes
 */
#define WRTD_GLOBAL_REP_CAP_ID "WGRCI"

/**
 * @file libwrtd-base.c
 */
/**
 * @defgroup base
 * Set of functions to manage the basic device and library configuration.
 * @{
 */
extern enum wrtd_status wrtd_init(const char *resource_name,
				  bool reset,
				  const char *options_str,
				  struct wrtd_dev **dev);
extern enum wrtd_status wrtd_close(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_reset(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_get_error(struct wrtd_dev *dev,
                                       enum wrtd_status *error_code,
                                       int32_t error_description_buffer_size,
                                       char *error_description);
extern enum wrtd_status wrtd_error_message(struct wrtd_dev *dev,
					   enum wrtd_status err_code,
					   char *err_message);


/* Attributes.  */
extern enum wrtd_status wrtd_get_attr_bool(struct wrtd_dev *dev,
					   const char *rep_cap_id,
					   enum wrtd_attr id,
					   bool *value);
extern enum wrtd_status wrtd_set_attr_bool(struct wrtd_dev *dev,
					   const char *rep_cap_id,
					   enum wrtd_attr id,
					   bool value);
extern enum wrtd_status wrtd_get_attr_int32(struct wrtd_dev *dev,
					    const char *rep_cap_id,
					    enum wrtd_attr id,
					    int32_t *value);
extern enum wrtd_status wrtd_set_attr_int32(struct wrtd_dev *dev,
					    const char *rep_cap_id,
					    enum wrtd_attr id,
					    int32_t value);
extern enum wrtd_status wrtd_get_attr_tstamp(struct wrtd_dev *dev,
					     const char *rep_cap_id,
					     enum wrtd_attr id,
					     struct wrtd_tstamp *value);
extern enum wrtd_status wrtd_set_attr_tstamp(struct wrtd_dev *dev,
					     const char *rep_cap_id,
					     enum wrtd_attr id,
					     const struct wrtd_tstamp *value);
extern enum wrtd_status wrtd_set_attr_string(struct wrtd_dev *dev,
					     const char *rep_cap_id,
					     enum wrtd_attr id,
					     const char *value);
extern enum wrtd_status wrtd_get_attr_string(struct wrtd_dev *dev,
					     const char *rep_cap_id,
					     enum wrtd_attr id,
					     int32_t value_buf_size,
					     char *value);

/**@}*/

/**
 * @file libwrtd-time.c
 */
/**
 * @defgroup time
 * Set of functions to manage time
 * @{
 */
extern enum wrtd_status wrtd_get_sys_time(struct wrtd_dev *dev,
					  struct wrtd_tstamp *time);

/**@}*/
extern enum wrtd_status wrtd_clear_event_log_entries(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_get_next_event_log_entry(struct wrtd_dev *wrtd,
                                                      int32_t log_entry_buffer_size,
                                                      char *log_entry);


/**
 * @file libwrtd-alarm.c
 */
/**
 * @defgroup alarms
 * Set of functions to configure alarms as
 * sources for the generation of events
 * @{
 */
extern enum wrtd_status wrtd_add_alarm(struct wrtd_dev *dev,
				       const char *rep_cap_id);
extern enum wrtd_status wrtd_disable_all_alarms(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_remove_alarm(struct wrtd_dev *dev,
					  const char *rep_cap_id);
extern enum wrtd_status wrtd_remove_all_alarms(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_get_alarm_id(struct wrtd_dev *dev,
					  int32_t index,
					  int32_t name_buffer_size,
					  char *rep_cap_id);

/**
 * @file libwrtd-rule.c
 */
/**
 * @defgroup rules
 * Set of functions to trigger on and
 * schedule the generation of events
 * @{
 */
// rule sources can be a) reserved event ids, b) alarm* c) any other string
// which will be interpreted as a net msg event id
extern enum wrtd_status wrtd_add_rule(struct wrtd_dev *dev,
				      const char *rep_cap_id);
extern enum wrtd_status wrtd_disable_all_rules(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_remove_rule(struct wrtd_dev *dev,
					 const char *rep_cap_id);
extern enum wrtd_status wrtd_remove_all_rules(struct wrtd_dev *dev);
extern enum wrtd_status wrtd_get_rule_id(struct wrtd_dev *dev,
					 int32_t index,
					 int32_t name_buffer_size,
					 char *rep_cap_id);
extern enum wrtd_status wrtd_reset_rule_stats(struct wrtd_dev *dev,
					      const char *rep_cap_id);


#ifdef __cplusplus
};
#endif

#endif
